
// Web Audio API check
window.AudioContext = window.AudioContext || window.webkitAudioContext;
if (!AudioContext) {
  alert("Sorry, your browser doesn't support the Web Audio APIs.");
  throw new Error("Sorry, your browser doesn't support the Web Audio APIs. Execution Aborted."); // ABORT ALL
}

let audio_ctx = new AudioContext();

let midikeyByKeyCode = new Map([
    //  G#2        A2       A#2        B2
    [65, 44], [90, 45], [83, 46], [88, 47],
    //   C3       C#3        D3       D#3        E3        F3       F#3        G3       G#3         A3       A#3         B3
    [67, 48], [70, 49], [86, 50], [71, 51], [66, 52], [78, 53], [74, 54], [77, 55], [75, 56], [188, 57], [76, 58], [190, 59],
    //   C4       C#4        D4       D#4        E4        F4       F#4        G4       G#4        A4       A#4        B4
    [81, 60], [50, 61], [87, 62], [51, 63], [69, 64], [82, 65], [53, 66], [84, 67], [54, 68], [89, 69], [55, 70], [85, 71],
    //   C5       C#5        D5       D#5        E5
    [73, 72], [57, 73], [79, 74], [48, 75], [80, 76]
]);

let keySymbolByMidikey = new Map([
    //   G#2         A2        A#2         B2
    [44, 'A'], [45, 'Z'], [46, 'S'], [47, 'X'],
    //    C3        C#3         D3        D#3         E3         F3        F#3         G3        G#3         A3        A#3         B3
    [48, 'C'], [49, 'F'], [50, 'V'], [51, 'G'], [52, 'B'], [53, 'N'], [54, 'J'], [55, 'M'], [56, 'K'], [57, ','], [58, 'L'], [59, '.'],
    //    C4        C#4         D4        D#4         E4         F4        F#4         G4        G#4         A4        A#4         B4
    [60, 'Q'], [61, '2'], [62, 'W'], [63, '3'], [64, 'E'], [65, 'R'], [66, '5'], [67, 'T'], [68, '6'], [69, 'Y'], [70, '7'], [71, 'U'],
    //    C5        C#5         D5        D#5         E5
    [72, 'I'], [73, '9'], [74, 'O'], [75, '0'], [76, 'P']
]);

const SCHEDULE_AHEAD_TIME = 1/10;

let notes = [];

let awpatch = new AWPatch();
let saved_awpatch = null;

pitchenv_view = new PixelCanvas(document.querySelector("#pitchenv-canvas"), 32, 32.0/32, awpatch.semitones, "#ffca28");
ampenv_view = new PixelCanvas(document.querySelector("#ampenv-canvas"), 16, 100.0/15, awpatch.amplitudes, "#42a5f5");

function load() {
    if(window.readFile) {
        let filePath = window.location.search.substr(1);
        awpatch.parse(window.readFile(filePath, "utf-8"));
    }
    
    waveformChanged(awpatch.waveform);
    
    document.getElementById("volume").value = awpatch.volume;
    document.getElementById("step").value = awpatch.step;
    document.getElementById("release").value = awpatch.release;
    document.getElementById("glide").value = awpatch.glide;
    
    this.pitchenv_view.setEnvelope(awpatch.semitones);
    document.getElementById("pitchenv_smooth").checked = awpatch.semitones.smooth;
    document.getElementById("pitchenv_loop_start").value = awpatch.semitones.loop_start;
    document.getElementById("pitchenv_length").value = awpatch.semitones.length;
    
    this.ampenv_view.setEnvelope(awpatch.amplitudes);
    document.getElementById("ampenv_smooth").checked = awpatch.amplitudes.smooth;
    document.getElementById("ampenv_loop_start").value = awpatch.amplitudes.loop_start;
    document.getElementById("ampenv_length").value = awpatch.amplitudes.length;
    
    pitchenv_view.draw();
    ampenv_view.draw();
    
    saved_awpatch = awpatch.copy();
}

function save() {
    if(!awpatch.match(saved_awpatch) && window.saveFile) {
        let filePath = window.location.search.substr(1);
        window.saveFile(filePath, JSON.stringify(awpatch));
        saved_awpatch = awpatch.copy();
    }
}

function waveformChanged(type) {
    awpatch.waveform = type;
    
    // Highlight the selected waveform icon
    document.querySelectorAll("#waveforms .item").forEach((elem, idx) => {
        elem.style.backgroundColor = (elem.children[0].id == type) ? "#555555" : "";
    });
}

function volumeChanged() {
    awpatch.volume = document.getElementById("volume").value;
}

function stepChanged() {
    awpatch.step = document.getElementById("step").value;
}

function releaseChanged() {
    awpatch.release = document.getElementById("release").value;
}

function glideChanged() {
    awpatch.glide = document.getElementById("glide").value;
}

function semitonesChanged() {
    let val = document.getElementById("pitchenv_smooth").checked;
    awpatch.semitones.smooth = val;
    
    val = document.getElementById("pitchenv_loop_start").value;
    awpatch.semitones.loop_start = parseInt(val);
    
    val = document.getElementById("pitchenv_length").value;
    awpatch.semitones.length = parseInt(val);
    
    pitchenv_view.draw();
}

function amplitudesChanged() {
    let val = document.getElementById("ampenv_smooth").checked;
    awpatch.amplitudes.smooth = val;
    
    val = document.getElementById("ampenv_loop_start").value;
    awpatch.amplitudes.loop_start = parseInt(val);
    
    val = document.getElementById("ampenv_length").value;
    awpatch.amplitudes.length = parseInt(val);
    
    ampenv_view.draw();
}

function createWhiteKey(midikey, slot) {
    let key = document.createElement("div");
    key.className = "white pianokey";
    key.style.height = "120px";
    key.style.width = "24px";

    key.dataset["midikey"] = midikey;

    return key;
}

function createBlackKey(midikey, slot) {
    let key = document.createElement("div");
    key.className = "black pianokey";
    key.style.height = "70px";
    key.style.width = "14px";

    key.dataset["midikey"] = midikey;

    return key;
}

function setupVirtualPiano(eventHandler) {
    const semitones = { white: [0,2,4,5,7,9,11], black: [1,3,6,8,10] };
    
    let keyboard = document.getElementById('pianokeyb');
    
    const idx_lo = 5;       // A0
    const idx_hi = 8*7 + 0; // C8
    let idx_white = idx_lo;
    while(idx_white <= idx_hi) {
        const octave = (idx_white/7) | 0;
        const midikey = 12 + 12*octave + semitones.white[idx_white%7];
        let key = createWhiteKey(midikey);
        key.style.left = 24*(idx_white-idx_lo) + "px";
        keyboard.appendChild(key);
        
        let label = document.createElement('div');
        label.className = 'label';
        if(keySymbolByMidikey.has(midikey)) label.innerHTML = "<b>"+keySymbolByMidikey.get(midikey)+"</b>";
        key.appendChild(label);
        
        key.addEventListener("mousedown", eventHandler, false);
        key.addEventListener("mouseup", eventHandler, false);
        key.addEventListener("mouseenter", eventHandler, false);
        key.addEventListener("mouseleave", eventHandler, false);
        
        ++idx_white;
    }
    
    let idx_black = 5 * ((idx_lo/7)|0);
    while(semitones.black[idx_black%5] < semitones.white[idx_lo%7]) {
        ++idx_black;
        if(idx_black%5 == 0) break;
    }
    
    const semitone_hi = 12*((idx_hi/7)|0) + semitones.white[idx_hi%7]
    while(12*((idx_black/5)|0) + semitones.black[idx_black%5] < semitone_hi) {
        const octave = (idx_black/5)|0;
        const midikey = 12 + 12*octave + semitones.black[idx_black%5];
        let key = createBlackKey(midikey);
        key.style.left = 14*(midikey-12) - 24*(idx_lo) + "px";
        keyboard.appendChild(key);
        
        let label = document.createElement('div');
        label.className = 'label';
        if(keySymbolByMidikey.has(midikey)) label.innerHTML = "<b>"+keySymbolByMidikey.get(midikey)+"</b>";
        key.appendChild(label);
        
        key.addEventListener("mousedown", eventHandler, false);
        key.addEventListener("mouseup", eventHandler, false);
        key.addEventListener("mouseenter", eventHandler, false);
        key.addEventListener("mouseleave", eventHandler, false);
        
        ++idx_black;
    }
    
    // Scroll C4 to middle
    document.querySelector("div[data-midikey='60']").scrollIntoView({ behavior:'auto', block:'center', inline:'center' });
}

function PixelCanvas(canvas, rowzero, scale, envelope, color) {
    this.canvas = canvas;
    this.w = this.canvas.width;
    this.h = this.canvas.height;
    this.rowzero = rowzero;
    this.scale = scale;
    this.envelope = envelope;
    this.color = color;
    
    let dpi = window.devicePixelRatio;
    let style_width = +window.getComputedStyle(this.canvas).getPropertyValue("width").slice(0, -2);
    let style_height = +window.getComputedStyle(this.canvas).getPropertyValue("height").slice(0, -2);
    this.canvas.setAttribute('width', style_width * dpi);
    this.canvas.setAttribute('height', style_height * dpi);
    
    this.ctx = this.canvas.getContext("2d");
    this.ctx.globalAlpha = 1;
    
    this.canvas.addEventListener("mousedown", this);
    this.canvas.addEventListener("mousemove", this);
}

PixelCanvas.prototype.handleEvent = function(evt) {
    if(evt.type == "mousedown" || (evt.type == "mousemove" && (evt.buttons&1))) {
        let rect = this.canvas.getBoundingClientRect();
        let x = evt.clientX - rect.left;
        x = this.canvas.width * x / this.canvas.clientWidth;
        let y = evt.clientY - rect.top;
        y = this.canvas.height * y / this.canvas.clientHeight;
        
        let step = ((this.w*x/this.canvas.width)/4)|0;
        if(step > this.envelope.data.length-1) step = this.envelope.data.length-1;
        let val = (this.h*y/this.canvas.height)|0;
        val = this.rowzero - ((val < 1) ? 1 : ((val < this.h-1) ? val : this.h-1));
        
        this.envelope.data[step] = (val*this.scale)|0;
        
        this.draw();
    }
}

PixelCanvas.prototype.setEnvelope = function(env) {
    this.envelope = env;
}

PixelCanvas.prototype.draw = function() {
    this.ctx.fillStyle = "#000000";
    this.ctx.fillRect(0, 0, this.canvas.width, this.canvas.height);
    
    let stem_color = "#446";
    for(let step=0; step<32; ++step) {
        let val = this.envelope.data[step] / this.scale;
        if(val > 0) {
            for(let j=0; j<=this.rowzero; ++j) {
                if(this.rowzero-j < val) {
                    this.pset(4*step+1, j, stem_color);
                    this.pset(4*step+2, j, stem_color);
                }
            }
        }
        else {
            for(let j=this.rowzero-1; j<this.h; ++j) {
                if(this.rowzero-j > (val+1)) {
                    this.pset(4*step+1, j, stem_color);
                    this.pset(4*step+2, j, stem_color);
                }
            }
        }
        let j = this.rowzero - val;
        this.pset(4*step+1, j, this.color);
        this.pset(4*step+2, j, this.color);
        this.pset(4*step+1, j-1, this.color);
        this.pset(4*step+2, j-1, this.color);
    }
    
    // Draw loop start and length markers
    for(let j=0; j<this.h; j+=2) {
        let step = this.envelope.loop_start;
        this.pset(4*this.envelope.loop_start, j, "#888");
        this.pset(4*this.envelope.length-1, j, "#888");
    }
}

PixelCanvas.prototype.pset = function(x, y, color) {
    x = +x;
    y = +y;
    
    const resX = this.canvas.width/this.w;
    let x0 = (x*resX)|0;
    let x1 = ((x+1)*resX)|0;
    
    const resY = this.canvas.height/this.h;
    let y0 = (y*resY)|0;
    let y1 = ((y+1)*resY)|0;
    
    if (x1 >= 0 && x0 < this.canvas.width && y1 >= 0 && y0 < this.canvas.height) {
        if(color) this.ctx.fillStyle = color;
        this.ctx.fillRect(x0, y0, x1-x0, y1-y0);
    }
}

function AWPatch() {
    this.waveform = "square";
    this.volume = 80;
    this.step = 5;
    this.release = 0;
    this.glide = 0;
    this.semitones = { smooth: true, loop_start: 32, length: 32, data: Array(32).fill(0) };
    this.amplitudes = { smooth: true, loop_start: 32, length: 32, data: Array(32).fill(100) };
}

AWPatch.prototype.parse = function(json) {
    if(!json) return;
    
    ob = JSON.parse(json);
    this.waveform = ob.waveform;
    this.volume = ob.volume;
    this.step = ob.step;
    this.release = ob.release;
    this.glide = ob.glide;
    this.semitones.smooth = ob.semitones.smooth;
    this.semitones.loop_start = ob.semitones.loop_start;
    this.semitones.length = ob.semitones.length;
    this.semitones.data = ob.semitones.data.map((x) => x);
    this.amplitudes.smooth = ob.amplitudes.smooth;
    this.amplitudes.loop_start = ob.amplitudes.loop_start;
    this.amplitudes.length = ob.amplitudes.length;
    this.amplitudes.data = ob.amplitudes.data.map((x) => x);
}

AWPatch.prototype.copy = function() {
    let other = new AWPatch();
    other.waveform = this.waveform;
    other.volume = this.volume;
    other.step = this.step;
    other.release = this.release;
    other.glide = this.glide;
    other.semitones.smooth = this.semitones.smooth;
    other.semitones.loop_start = this.semitones.loop_start;
    other.semitones.length = this.semitones.length;
    other.semitones.data = this.semitones.data.map((x) => x);
    other.amplitudes.smooth = this.amplitudes.smooth;
    other.amplitudes.loop_start = this.amplitudes.loop_start;
    other.amplitudes.length = this.amplitudes.length;
    other.amplitudes.data = this.amplitudes.data.map((x) => x);
    return other;
}

AWPatch.prototype.match = function(other) {
    return this.waveform == other.waveform && this.volume == other.volume &&
        this.step == other.step && this.release == other.release && this.glide == other.glide &&
        this.semitones.smooth == other.semitones.smooth &&
        this.semitones.loop_start == other.semitones.loop_start &&
        this.semitones.length == other.semitones.length &&
        this.semitones.data.every((val,idx) => val === other.semitones.data[idx]) &&
        this.amplitudes.smooth == other.amplitudes.smooth &&
        this.amplitudes.loop_start == other.amplitudes.loop_start &&
        this.amplitudes.length == other.amplitudes.length &&
        this.amplitudes.data.every((val,idx) => val === other.amplitudes.data[idx]);
}

// Takes logarithmic level (0 to 1.0) and returns gain
function levelToGain(level) {
    return level > 0 ? 2**(-(1.0-level)*3.32193*35/20) : 0; // 3.32193=log2(10); 35=scales level to -35...0 dB
}

function initNote(midikey) {
    let note = {
        gain_node: audio_ctx.createGain(),
        osc: audio_ctx.createOscillator(),
        midikey: midikey,
        volume: awpatch.volume/100.0,
        
        release_rate: awpatch.release > 0 ? (awpatch.volume/100.0)/(awpatch.release*awpatch.step) : awpatch.volume/100.0,

        glide_interval: 0,
        glide_accu: 0,
        glide_rate: 0,
        
        step_accu: 0,
        step_rate: 1.0/awpatch.step + 1e-18,
        
        base_level: awpatch.amplitudes.data[0]/100.0,
        delta_level: awpatch.amplitudes.smooth ? (awpatch.amplitudes.data[1]-awpatch.amplitudes.data[0])/100.0 : 0,
        
        base_pitchbend: awpatch.semitones.data[0]/12,
        delta_pitchbend: awpatch.semitones.smooth ? (awpatch.semitones.data[1]-awpatch.semitones.data[0])/12 : 0,
        
        amplitude_idx: 1,
        semitone_idx: 1,
        
        last_scheduled_step_time: audio_ctx.currentTime+0.05,
        
        released: false
    };
    
    note.gain_node.connect(audio_ctx.destination);
    note.gain_node.gain.setValueAtTime(0, 0);
    let level = note.volume * (note.base_level);
    note.gain_node.gain.setValueAtTime(levelToGain(level), note.last_scheduled_step_time);
    
    note.osc.connect(note.gain_node);
    note.osc.type = awpatch.waveform;
    let pitch = 440 * 2**((note.midikey-69)/12);
    pitch *= 2**(note.base_pitchbend);
    note.osc.frequency.setValueAtTime(pitch, 0);
    
    return note;
}

function update(note) {
    let step_time = note.last_scheduled_step_time + 1.0/120;
    while(step_time < audio_ctx.currentTime+SCHEDULE_AHEAD_TIME && note.volume > 0.0) {
        note.step_accu += note.step_rate;
        if(note.step_accu >= 1.0) {
            note.step_accu = 0.0;
            
            const amplitudes = awpatch.amplitudes;
            let len = (amplitudes.length > 0 && amplitudes.length < 32) ? amplitudes.length : 32;
            if(note.amplitude_idx < len) {
                note.base_level = amplitudes.data[note.amplitude_idx]/100.0;
                
                ++note.amplitude_idx;
                if(note.amplitude_idx >= len) {
                    if(amplitudes.loop_start < amplitudes.length) {
                        note.amplitude_idx = amplitudes.loop_start;
                        note.delta_level = amplitudes.smooth ? (amplitudes.data[note.amplitude_idx]/100.0 - note.base_level) : 0;
                    }
                    else {
                        note.amplitude_idx = 32;
                        note.delta_level = 0;
                    }
                }
                else {
                    note.delta_level = amplitudes.smooth ? (amplitudes.data[note.amplitude_idx]/100.0 - note.base_level) : 0;
                }
            }
            else if(!note.released) {
                note.released = true;
            }

            const semitones = awpatch.semitones;
            len = (semitones.length > 0 && semitones.length < 32) ? semitones.length : 32;
            if(note.semitone_idx < len) {
                note.base_pitchbend = semitones.data[note.semitone_idx]/12; // Convert semitone to octave
                
                ++note.semitone_idx;
                if(note.semitone_idx >= len) {
                    note.semitone_idx = semitones.loop_start < semitones.length ? semitones.loop_start : (len-1);
                }
                
                note.delta_pitchbend = semitones.smooth ? (semitones.data[note.semitone_idx]/12 - note.base_pitchbend) : 0;
            }
        }
        
        if(note.released) {
            note.volume -= note.release_rate;
            if(note.volume <= 0.0) {
                note.volume = 0.0;
            }
        }
        
        let level = note.volume * (note.base_level + note.delta_level*note.step_accu);
        note.gain_node.gain.linearRampToValueAtTime(levelToGain(level), step_time);
        
        let pitchbend = note.base_pitchbend + note.delta_pitchbend*note.step_accu;

        if(note.glide_accu > 0) {
            note.glide_accu -= note.glide_rate;
            if(note.glide_accu < 0) {
                note.glide_accu = 0;
            }
            else {
                pitchbend += (note.glide_interval*note.glide_accu);
            }
        }
        let pitch = 440 * 2**((note.midikey-69)/12);
        note.osc.frequency.setValueAtTime(pitch * 2**(pitchbend), step_time);
        
        note.last_scheduled_step_time = step_time;
        step_time += 1/120;
    }
    
    if(note.last_scheduled_step_time >= audio_ctx.currentTime) {
        window.setTimeout(update, 1/120, note);
    }
    else {
        note.osc.stop(0);
        note.osc.disconnect();
        note.gain_node.disconnect();
    }
}

function noteOn(midikey) {
    if(awpatch.glide > 0 && notes.length > 0 && !notes[0].released) {
        let note = notes[0];
        if(note.midikey != midikey) {
            note.glide_interval = note.glide_interval*note.glide_accu + (note.midikey-midikey)/12;
            note.glide_accu = 1.0;
            note.glide_rate = 1.0/(awpatch.glide*awpatch.step);
            
            note.midikey = midikey;
        }
    }
    else {  
        let alreadyPressed = false;
        for(let i = 0; i < notes.length; i++) {
            if(notes[i].midikey == midikey) {
                alreadyPressed = true;
                break;
            }
        }
        if(!alreadyPressed) {
            let note = initNote(midikey);
            update(note);
            note.osc.start(audio_ctx.currentTime);
            notes.push(note);
        }
    }
}

function noteOff(midikey) {
    let garbage = [];
    for(let i = 0; i < notes.length; i++) {
        if(notes[i].midikey == midikey) {
            notes[i].released = true;
            garbage.push(i);
        }
    }
    for(let i = 0; i < garbage.length; i++) {
        notes.splice(garbage[i], 1);
    }
}

function virtualPianoEventHandler(event) {
    if(event.type == "mousedown" && (event.buttons & 1)) {
        noteOn(event.target.dataset["midikey"]);
    }
    if(event.type == "mouseenter" && (event.buttons & 1)) {
        noteOn(event.target.dataset["midikey"]);
        // If entering from another key, turn the previous note off
        if(event.relatedTarget.classList.contains("pianokey")) {
            noteOff(event.relatedTarget.dataset["midikey"]);
        }
    }
    else if(event.type == "mouseup") {
        noteOff(event.target.dataset["midikey"]);
    }
    else if(event.type == "mouseleave" && (event.buttons & 1)) {
        // Don't turn note off when moving from one piano key to another, only when moving entirely outside the keyboard
        if(!event.relatedTarget.classList.contains("pianokey")) {
            noteOff(event.target.dataset["midikey"]);
        }
    }
}

document.addEventListener('keydown', function (event) {
    if(midikeyByKeyCode.has(event.keyCode)) {
        noteOn(midikeyByKeyCode.get(event.keyCode));
    }
}, false);

document.addEventListener('keyup', function (event) {
    if(midikeyByKeyCode.has(event.keyCode)) {
        noteOff(midikeyByKeyCode.get(event.keyCode));
    }
}, false);

window.onload = function (event) {
    setupVirtualPiano(virtualPianoEventHandler);    // Create the virtual piano keyboard
    window.scroll(0,0);                             // Scroll back to top, for some reason FemtoIDE always scrolls to bottom
    
    load();                         // Load AWPatch from file
    
    window.setInterval(save, 1000); // Setup auto-save every 1 seconds
}

